---------------------------------------------------------------------------------------
-- This file is a adder module, it performs addition of std_logic_vector multiple 
-- clock cycles. The number of clock cycles taken depend upon adder_width

-- Copyright (C) 2021
-- Authors: Sanjay Deshpande <sanjay.deshpande@yale.edu>
          
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 3 of the License, or
-- (at your option) any later version.

-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.

-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software Foundation,
-- Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
--
--------------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_unsigned.ALL;

entity adder is
GENERIC (
		w : integer := 741; -- width of the input
		adder_width : integer := 247 -- width of the adder

		);
		
		PORT  (
          clk : in std_logic;
          start : in std_logic;
          done : out std_logic;
          a_in : in std_logic_vector(w-1 downto 0); 
          b_in : in std_logic_vector(w-1 downto 0); 
          c_out : out std_logic_vector(w-1 downto 0)
        );
end adder;

architecture Behavioral of adder is
signal c_out_temp : std_logic_vector(w - 1 downto 0);
signal a_in_reg : std_logic_vector(w - 1 downto 0);
signal b_in_reg : std_logic_vector(w - 1 downto 0);
signal c_out_reg : std_logic_vector(w-1 downto 0) := (others => '0');
signal c0 : std_logic_vector(adder_width downto 0) := (others => '0');
signal c1 : std_logic_vector(adder_width-1 downto 0) := (others => '0');
signal c2 : std_logic_vector(adder_width-1 downto 0) := (others => '0');
signal done_reg : std_logic_vector(w/adder_width+1 downto 0) := (others => '0');
constant ZEROS : std_logic_vector(adder_width-1 downto 0) := (others => '0');
signal carry : std_logic := '0';
begin

process(clk) 
begin
	if rising_edge(clk) then
		if start = '1' then
			a_in_reg <= a_in;
			b_in_reg <= b_in;
		else
			a_in_reg <= zeros & a_in_reg(w-1 downto adder_width);
			b_in_reg <= zeros & b_in_reg(w-1 downto adder_width);
		end if;
	end if;
end process;

process(clk) 
begin
	if rising_edge(clk) then
		if start = '1' then
			carry <= '0';
		else
			carry <= c0(c0'high);
		end if;
	end if;
end process;

c0 <= '0'&a_in_reg(adder_width -1 downto 0) + b_in_reg(adder_width -1 downto 0) + carry;

process(clk) 
begin
	if rising_edge(clk) then
	   if start = '1' then
    		c_out_reg <= (others => '0');
       else
            c_out_reg <=c0(adder_width - 1 downto 0) & c_out_reg(c_out_reg'high downto adder_width);
       end if; 
	end if;
end process;

c_out <= c_out_reg;

done_reg(0) <= start;

done_track: for i in 0 to w/adder_width generate
                onereg: entity work.reg_one_bit(behavioral)
                        port map (
                                    clk => clk,
                                    d => done_reg(i),
                                    q => done_reg(i+1)
                        );
            end generate;

done <= done_reg(w/adder_width+1);

end Behavioral;



