--------------------------------------------------------------------------------------------
-- This file is a fast_gcd top module  

-- Copyright (C) 2021
-- Authors: Sanjay Deshpande <sanjay.deshpande@yale.edu>
          
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 3 of the License, or
-- (at your option) any later version.

-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.

-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software Foundation,
-- Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA


-- Note: Use this file only for testing purpose, 
--       This module is not optimized because existing modular multiplication and 
--       modular reduction from the system can be used,
--       For testing purpose we use simple * for multiplication and a full combinatorial 
--       Barrett reduction from the book Hardware Implementation of Finite-Field Arithmetic
--       By Jean Pierre Deschamps, Jose Luis Imana, and Gustavo D. Sutter
--------------------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use work.gcd_pkg.all;

entity fast_gcd is
  Port (
         clk          : in std_logic;
         rst        : in std_logic;
         start        : in std_logic;
         done         : out std_logic;
         f_in         : in STD_LOGIC_VECTOR (K-1 downto 0); 
         g_in         : in STD_LOGIC_VECTOR (K-1 downto 0);
         inv_out      : out STD_LOGIC_VECTOR(K-1 downto 0) 
         );
end fast_gcd;

architecture Behavioral of fast_gcd is

signal f_out    : std_logic_vector(K-1 downto 0);
signal M2Q_v_out    : std_logic_vector(t downto 0);
signal v_shift_count    : std_logic_vector(logt-1 downto 0);
signal two_power_t_1_shifted  : std_logic_vector(t downto 0) := (others => '0'); 
signal V  : std_logic_vector(2*t+3+ 1 -1 downto 0);
signal V_shifted  : std_logic_vector(2*t+3+ 1 -1 + 2 downto 0);
signal final_out  : std_logic_vector(3*t+3+K-1 downto 0) := (others=> '0');
signal final_out_red  : std_logic_vector(K-1 downto 0);
signal shift_int_v: integer range 0 to 2*t+t+1;
constant minus_1  : std_logic_vector(1 downto 0) :=(others => '1');
constant plus_1  : std_logic_vector(1 downto 0) :=(0=> '1',others => '0');
signal mul_one  : std_logic_vector(1 downto 0) :=(0=> '1',others => '0');
signal mul_one_v  : std_logic_vector(1 downto 0) :=(0=> '1',others => '0');

begin
   divstep: entity work.divsteps2(Behavioral) 
        GENERIC MAP(
            K => K, 
            T => T, 
            logT => logt,
            K_adder_width => K_adder_width, 
            t_adder_width => t_adder_width,
            K_adder_slices => K_adder_slices,
            t_adder_slices => t_adder_slices,
            barrel_shift_width => barrel_shift_width
        )
        PORT MAP (
          clk => clk,
          rst => rst,
          start => start,
          done => done,
          n_in => n_in,
          t_in => t_in,
          delta_in =>(0=> '1', others => '0'),
          f_in => f_in,
          g_in => g_in,
          f_out => f_out,
          M2Q_v_out => M2Q_v_out,
          v_shift_count => v_shift_count
        );



t_shift: for position in two_power_t_1'range generate
            two_power_t_1_shifted(position) <= two_power_t_1((position + shift_int_v) mod V'high) when ((position  < two_power_t_1'high - shift_int_v)) else '0';
        end generate t_shift;

V <=  std_logic_vector(signed(M2Q_v_out(t downto 0)) * signed(two_power_t_1_shifted(t downto 0)) * signed(mul_one));  
shift_int_v <= to_integer(unsigned(v_shift_count));
mul_one <= plus_1 when f_out(f_out'high) = '0' else minus_1;
mul_one_v <= plus_1 when V(V'high) = '0' else minus_1;
v_shifted <= std_logic_vector(signed(V) * signed(mul_one_v));
final_out(2*t+1+3+K + 2 -1 downto 0) <= std_logic_vector(unsigned(V_shifted)*unsigned(precomp));

  barrett: entity work.Barrett_reducer_comb(rtl) PORT MAP (
          x => final_out,
          z => final_out_red(K-1 downto 0)
        );

inv_out <= std_logic_vector(unsigned(m(K-1 downto 0)) - unsigned(final_out_red(K-1 downto 0))) when V(V'high) = '1' else final_out_red;
        
end Behavioral;
