-------------------------------------------------------------------------------------
-- This file is  default elgamal decryption module which uses modular exponentiation 
-- for inverse 

-- Copyright (C) 2021
-- Authors: Sanjay Deshpande <sanjay.deshpande@yale.edu>
          
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 3 of the License, or
-- (at your option) any later version.

-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.

-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software Foundation,
-- Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

-- Psuedocode for ElGamal Decryption
--Input: Ciphertext 1 (c1_input), Ciphertext 2 (c2_input), SecretKey (x_input)
--Output: Message (output)
--Step 1: Compute s := c1_inputx
--Step 2: Compute ell := s(-1)
--Step 3: Compute output := (c2_input).ell

--parameter details
-- traditional_dsp = '1' uses the traditional dsp (add ip/add_dsp.xci and 
-- ip/mul_and_acc_dsp.xci to the project)

-- traditional_dsp = '0' uses vhdl-code and infers dsps (add add.vhd and 
--  mul_and_acc.vhd to the project) 


-- parameters ramsize and addr_width are fixed values as the design is targeted to 
-- specific prime (2^1279 -1)
-------------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_unsigned.ALL;
use IEEE.NUMERIC_STD.ALL;



entity elgamal_decryption is
generic (
            traditional_dsp : std_logic := '1';
            
            
            ramsize : integer := 80;    
            addr_width : integer := 7
        );
Port ( 
           clk   : in STD_LOGIC;
           start : in STD_LOGIC;
           rst   : in STD_LOGIC;
           done  : out STD_LOGIC;
           valid  : out STD_LOGIC;
           c1_input : in STD_LOGIC_VECTOR (15 downto 0);
           c2_input : in STD_LOGIC_VECTOR (15 downto 0);
           x_input : in STD_LOGIC_VECTOR (15 downto 0);
           output: out STD_LOGIC_VECTOR (15 downto 0)

           );
end elgamal_decryption;

architecture Behavioral of elgamal_decryption is
    COMPONENT mul_and_acc_dsp
    PORT(
         clk : in STD_LOGIC;
           a : in STD_LOGIC_VECTOR (17 downto 0);
           b : in STD_LOGIC_VECTOR (17 downto 0);
           c : in STD_LOGIC_VECTOR (47 downto 0);
           p : out STD_LOGIC_VECTOR (47 downto 0)  
        );
    END COMPONENT;
    
    COMPONENT add_dsp
    PORT(
         clk : in STD_LOGIC;
           CONCAT : in STD_LOGIC_VECTOR (47 downto 0);
           C : in STD_LOGIC_VECTOR (47 downto 0);
           p : out STD_LOGIC_VECTOR (47 downto 0);
           CARRYIN : in STD_LOGIC  
--           CARRYOUT : out STD_LOGIC  
        );
    END COMPONENT;
    


signal a_addr_a, a_addr_b : std_logic_vector(addr_width-1 downto 0) := (others => '0');
signal a_wea, a_web: std_logic := '0';
signal a_ram_in_a, a_ram_in_b : std_logic_vector(15 downto 0);    
signal a_ram_out_a, a_ram_out_b : std_logic_vector(15 downto 0); 

signal b_addr_a, b_addr_b : std_logic_vector(addr_width-1 downto 0) := (others => '0');
signal b_wea, b_web: std_logic := '0';
signal b_ram_in_a, b_ram_in_b : std_logic_vector(15 downto 0);    
signal b_ram_out_a, b_ram_out_b : std_logic_vector(15 downto 0);  



signal mul_addr_a, mul_addr_b : std_logic_vector(addr_width downto 0) := (others => '0');
signal t_addr_a, t_addr_b, t_addr_b_reg : std_logic_vector(addr_width downto 0) := (others => '0');
signal t_wea, t_web: std_logic := '0';
signal t_ram_in_a, t_ram_in_b : std_logic_vector(15 downto 0);    
signal t_ram_out_a, t_ram_out_b, t_ram_out_b_op : std_logic_vector(15 downto 0);

signal a,b : std_logic_vector(17 downto 0);
signal c,p : std_logic_vector(47 downto 0);
signal adder_in_1, adder_in_2 : std_logic_vector(47 downto 0);
signal dsp_add_out : std_logic_vector(47 downto 0) := (others=> '0');
signal carryin_adder : std_logic := '0'; 

signal p_reg : std_logic_vector(31 downto 0) := (others => '0');
signal p_reg_reg : std_logic_vector(31 downto 0) := (others => '0');

signal p_op : std_logic_vector(15 downto 0) := (others => '0');

signal no_add : std_logic := '0';
signal first_add : std_logic := '0';

signal valid_s : std_logic := '0';
signal done_s : std_logic := '0';


--test
signal load_done_t : std_logic := '0';
signal carryout : std_logic := '0';

signal count, count_reg, count_reg_reg : integer range 0 to 4*ramsize + 1;
signal count_add : integer range 0 to 4*ramsize + 1;


type state_s is (S_RESET, S_WAIT_START, S_STALL_0, S_INIT_LOAD, S_LOAD, S_STALL_1, S_INC_A, S_INC_B, S_STALL_2, S_DONE);
signal state        			: state_s;  

type state_s2 is (S_RESET_add, S_WAIT_count, S_ADD_1,S_ADD_2, S_ADD_3, S_ADD_4, S_STALL_ADD_1, S_ADDING, S_MUL_to_RED, S_RED_STALL_0, S_RED_STALL_1, S_RED, S_RED_2, S_RED_STALL_2, S_ALL_ONES, S_READ_OUT);
signal state_add        			: state_s2; 




constant zeros : std_logic_vector(47 downto 0) := (others =>'0');


--=============== Reduction ===========================
signal t_ram_out_b_reg : std_logic; 
signal red_add_in_mux_1 : std_logic_vector(15 downto 0) := (others => '0');
signal red_add_in_mux_2 : std_logic_vector(15 downto 0) := (others => '0');
signal red_add_in_mux_3 : std_logic_vector(15 downto 0) := (others => '0');

signal red_adder_in_1 : std_logic_vector(47 downto 0) := (others => '0'); 
signal red_adder_in_2 : std_logic_vector(47 downto 0) := (others => '0'); 
signal red_add_out : std_logic_vector(47 downto 0) := (others => '0'); 

signal second_round : std_logic := '0'; 
signal first_add_red : std_logic := '0'; 
signal second_add_red : std_logic := '0'; 

signal sq_reg  : std_logic := '0'; 
signal reduction  : std_logic := '0'; 
signal carry_in : std_logic := '0'; 
signal last_block, last_block_red_2 : std_logic := '0';

signal b_addr_a_mux, b_addr_b_mux : std_logic_vector(addr_width-1 downto 0) := (others => '0');
signal b_wea_mux, b_web_mux: std_logic := '0';

signal red_b_addr_a, red_b_addr_b : std_logic_vector(addr_width-1 downto 0) := (others => '0');
signal red_b_wea, red_b_web: std_logic := '0';
signal red_b_ram_in_a, red_b_ram_in_b : std_logic_vector(15 downto 0);    
signal red_b_ram_out_a, red_b_ram_out_b : std_logic_vector(15 downto 0); 
signal count_all_ones : integer range 0 to ramsize;


-- Exponentiation Signals
 
type state_s3 is (S_RESET_exp, S_WAIT_exp, S_INIT_LOAD_exp, S_STALL_FOR_LOAD_E, S_LOAD_E_OUT_exp, S_SQUARE_exp, S_WAIT_SQUARE_DONE_exp, S_MUL_exp, S_WAIT_MUL_DONE_exp);
signal state_exp        			: state_s3; 
signal exp_a_addr_a, exp_b_addr_a : std_logic_vector(addr_width-1 downto 0) := (others => '0');
signal exp_a_wea, exp_b_wea, exp_b_web : std_logic:=  '0';
signal a_addr_a_mux, a_addr_b_mux : std_logic_vector(addr_width-1 downto 0) := (others => '0');
signal a_wea_mux, a_web_mux: std_logic := '0';
signal exp : std_logic := '0';
signal count_exp : integer range 0 to 16;
signal e_reg : std_logic_vector(15 downto 0) := (others => '0');
signal exp_shift : std_logic := '0';
signal exp_init, first_exp_init : std_logic := '0';
signal start_mod_mul, start_mod_mul_exp : std_logic := '0';
signal done_mod_mul : std_logic := '0';
signal e_mod_2 : std_logic := '0';
signal r_initial : std_logic := '0';
signal first_block_exp : std_logic := '0';

signal e_addr_a, e_addr_b : std_logic_vector(addr_width-1 downto 0) := (others => '0');
signal e_addr_count : std_logic_vector(addr_width downto 0) := (others => '0');
signal e_wea, e_web: std_logic := '0';
signal e_ram_in_a, e_ram_in_b : std_logic_vector(15 downto 0);    
signal e_ram_out_a, e_ram_out_b : std_logic_vector(15 downto 0); 

signal start_expo, done_expo : std_logic := '0'; 

constant ones : std_logic_vector(e_addr_count'high downto 0) := (others =>'1');

-- Decryption Signals

type state_s4 is (S_RESET_dec, S_WAIT_dec, S_INIT_LOAD_dec, S_C1PX, S_C1PX_done, S_MOV_B2A, S_MOV_B2A_rem, S_C1PX_INV, S_C1PX_INV_done, S_MOV_C2A, S_MOV_C2A_rem, S_MUL_C1XC2,S_MUL_C1XC2_done, S_READ_OUT_dec);
signal state_dec        			: state_s4; 


signal c_addr_a, c_addr_b : std_logic_vector(addr_width-1 downto 0) := (others => '0');
signal c_wea, c_web: std_logic := '0';

signal c_ram_in_a, c_ram_in_b : std_logic_vector(15 downto 0);    
signal c_ram_out_a, c_ram_out_b, c_ram_out_b_op : std_logic_vector(15 downto 0);

signal e_wea_mux, e_web_mux: std_logic := '0';
signal e_addr_a_mux, e_addr_b_mux : std_logic_vector(addr_width-1 downto 0) := (others => '0');

signal cb_addr_b, ca_addr_b : std_logic_vector(addr_width-1 downto 0) := (others => '0');
signal cb_addr_a, ca_addr_a : std_logic_vector(addr_width-1 downto 0) := (others => '0');
signal ca_wea, cb_wea: std_logic := '0';
signal ca_web, cb_web: std_logic := '0';

signal ec_addr_a, ec_addr_b : std_logic_vector(addr_width-1 downto 0) := (others => '0');
signal ec_wea, ec_web: std_logic := '0';


signal dec: std_logic := '0';
signal move_b2a, sel_c2a: std_logic := '0';

signal start_mul_dec: std_logic := '0';
signal valid_output, done_dec: std_logic := '0';

begin

output <= b_ram_out_a(15 downto 0);


e_ram_in_a <= x"7fff" when e_addr_a_mux = ramsize -1 else
              x"fffd" when e_addr_a_mux = 0 else
              x"ffff";
e_ram_in_b <= x_input;

e_addr_a_mux <= ec_addr_a when move_b2a ='1' else e_addr_a(ADDR_WIDTH-1 downto 0);
e_addr_b_mux <= ec_addr_b when dec ='1' else e_addr_b(ADDR_WIDTH-1 downto 0);

e_wea_mux <= ec_wea when move_b2a = '1' else e_wea;
e_web_mux <= ec_web when dec = '1' else e_web;

	mem_e: entity work.memram(memory) 
            generic map(
                            SIZE => ramsize,
                            ADDR_WIDTH => addr_width,
                            COL_WIDTH => 16
                        )
            port map(
                        clk => clk,
                        wea => e_wea_mux, 
                        web => e_web_mux,  
                        address_a => e_addr_a_mux(ADDR_WIDTH-1 downto 0),
                        address_b => e_addr_b_mux(ADDR_WIDTH-1 downto 0),
                        datain_a => e_ram_in_a,
                        datain_b => e_ram_in_b,
                        dataout_a => e_ram_out_a,
                        dataout_b  => e_ram_out_b         
            );

a_ram_in_a <= c_ram_out_b when sel_c2a = '1' else  b_ram_out_a;
a_ram_in_b <= c1_input when dec = '1' else (others => '0');

a_addr_a_mux <= exp_a_addr_a when exp = '1' else 
                ca_addr_a when move_b2a = '1' else
                a_addr_a;
a_addr_b_mux <= ca_addr_b when dec = '1' else a_addr_b;

a_wea_mux <= exp_a_wea when exp = '1' else 
             ca_wea when move_b2a = '1' else
             a_wea; 

a_web_mux <= ca_web when dec = '1' else a_web; 

	mem_a: entity work.memram(memory) 
            generic map(
                            SIZE => ramsize,
                            ADDR_WIDTH => addr_width,
                            COL_WIDTH => 16
                        )
            port map(
                        clk => clk,
                        wea => a_wea_mux, 
                        web => a_web_mux,  
                        address_a => a_addr_a_mux,
                        address_b => a_addr_b_mux,
                        datain_a => a_ram_in_a,
                        datain_b => a_ram_in_b,
                        dataout_a => a_ram_out_a,
                        dataout_b  => a_ram_out_b         
            );
 
 

r_initial <= '1' when (dec ='1' or move_b2a = '1') and cb_addr_b = 0 else '0';
b_ram_in_a <= red_b_ram_in_a;
b_ram_in_b <= "000000000000000" & r_initial when (dec = '1' or move_b2a = '1') else (others => '0');

b_addr_a_mux <= red_b_addr_a when reduction = '1' else 
                cb_addr_a when move_b2a = '1' else
                b_addr_a; 
                
b_addr_b_mux <=  red_b_addr_b when reduction = '1' else 
                 cb_addr_b when (dec = '1' or move_b2a = '1') else
                 b_addr_b; 

b_wea_mux <= '0' when e_mod_2 = '1' else
             red_b_wea when reduction ='1' else 
             b_wea;

b_web_mux <= red_b_web when reduction = '1' else 
             cb_web when (dec = '1' or move_b2a = '1') else
             b_web;

    
     mem_b: entity work.memram(memory) 
            generic map(
                            SIZE => ramsize,
                            ADDR_WIDTH => addr_width,
                            COL_WIDTH => 16
                        )
            port map(
                        clk => clk,
                        wea => b_wea_mux, 
                        web => b_web_mux,  
                        address_a => b_addr_a_mux,
                        address_b => b_addr_b_mux,
                        datain_a => b_ram_in_a,
                        datain_b => b_ram_in_b,
                        dataout_a => b_ram_out_a,
                        dataout_b  => b_ram_out_b         
            );
            
 


 a <= "00" & a_ram_out_b when sq_reg = '0' else "00" & b_ram_out_a;
 b <= "00" & b_ram_out_b;
 
 c <= (others => '0') when b_addr_b = 1 else x"00000000"&p(31 downto 16);

--====================MUL and ACC=============================
trad_dsp:if traditional_dsp = '1' generate          
uut_mulacc: mul_and_acc_dsp PORT MAP (
          clk => clk,
          a => a,
          b => b,
          c => c,
          p => p
        );
end generate trad_dsp;

infer_dsp:if traditional_dsp = '0' generate    
uut_mulacc_2: entity work.mul_and_acc(behavioral) PORT MAP (
          clk => clk,
          a => a,
          b => b,
          c => c,
          p => p
        );
end generate infer_dsp;
--==============================================================

p_register:process(clk)
begin
    if rising_edge(clk) then
        p_reg <= p(31 downto 0);
        p_reg_reg <= p_reg;
    end if;
end process;

t_ram_out_b_op <= t_ram_out_b when no_add ='0' else (others => '0');
p_op <= p(15 downto 0) when no_add ='0' else p_reg(31 downto 16);
adder_in_1 <= zeros(47 downto t_ram_in_b'high + 1) & t_ram_out_b_op;
adder_in_2 <=  zeros(47 downto t_ram_in_b'high + 1) & p_op(15 downto 0);
carryin_adder <= '0' when first_add = '1' else dsp_add_out(16);

--=========================ADD 1==================================
trad_dsp_add:if traditional_dsp = '1' generate 
uut_add: add_dsp PORT MAP (
          clk => clk,
          CONCAT => adder_in_1,
          C => adder_in_2,
          CARRYIN => carryin_adder,
          p => dsp_add_out
        );
end generate trad_dsp_add;

infer_dsp_add:if traditional_dsp = '0' generate 
uut_add: entity work.add(behavioral) PORT MAP (
          clk => clk,
          CONCAT => adder_in_1,
          C => adder_in_2,
          CARRYIN => carryin_adder,
          p => dsp_add_out
        );
end generate infer_dsp_add;
--==============================================================
        
t_ram_in_a <= p(15 downto 0) when count_add <= (ramsize-1) and no_add = '1' else 
              p_reg(31 downto 16) when count_add = ramsize and no_add = '1' else
              dsp_add_out(15 downto 0);

t_ram_in_b <= p_reg(31 downto 16); 

 mem_temp: entity work.memram(memory) 
        generic map(
                        SIZE => 2*ramsize,
                        ADDR_WIDTH => addr_width + 1,
                        COL_WIDTH => 16
                    )
        port map(
                    clk => clk,
                    wea => t_wea, 
                    web => t_web,  
                    address_a => t_addr_a,
                    address_b => t_addr_b,
                    datain_a => t_ram_in_a,
                    datain_b => t_ram_in_b,
                    dataout_a => t_ram_out_a,
                    dataout_b  => t_ram_out_b         
        );

 start_mod_mul <= start_mod_mul_exp or start_mul_dec;       
p_fsm: process(clk)
    begin
        if rising_edge(clk) then
            if (rst = '1') then   
            else
				if (state = S_RESET) then
					state           <= S_WAIT_START;
					
				elsif (state = S_WAIT_START) then
			         if start_mod_mul = '1' then
			             state <= S_STALL_1;
			             a_addr_a <= (others => '0');
			             b_addr_a <= (others => '0');
			             a_addr_b <= (others => '0');
			             b_addr_b <= (others => '0');			             		             
			         end if;
			    
				
			   elsif (state = S_STALL_1) then
			         state <= S_INC_B;
			         count <= 0;
			         
			    elsif (state = S_INC_B) then 
			         count <= count + 1;
                     if (b_addr_b = ramsize -1)  then                          
                         b_addr_b <= (others => '0');
                         
                         if a_addr_b =  ramsize - 1 then
                            a_addr_b <= (others => '0');
                            b_addr_a <= (others => '0');
                            state <= S_WAIT_START;
							count <= 0;
                         else                           
                            state <= S_INC_A;
                         end if;  
                    else
                         b_addr_b <= b_addr_b + 1;
                     end if;
                
                elsif (state = S_INC_A) then
                     state <= S_INC_B;
                     a_addr_b <= a_addr_b + 1;
                     b_addr_a <= b_addr_a + 1;
                  
			end if;
			   end if;
          
         count_reg <= count;
         count_reg_reg <=count_reg;
         end if;
    end process;
    
	
    p_comb: process(state, start, a_addr_a, a_addr_b, b_addr_a, b_addr_b, count)
    begin
        --! Default value
        a_wea <= '0';
        a_web <= '0';
        b_wea <= '0';
        b_web <= '0';

        
      if (state = S_RESET) then

		elsif (state = S_WAIT_START) then

		    		
		elsif (state = S_INIT_LOAD) then
	        a_wea <= '1';
		    b_wea <= '1';       
		   
		elsif (state = S_INC_B) then

		
		elsif (state = S_STALL_2) then
		  	  
        end if;
		end process;	


p_fsm_add: process(clk)
    begin
        if rising_edge(clk) then
            if (rst = '1') then   
            else
				if (state_add = S_RESET_add) then
					state_add    <= S_WAIT_count;
					t_addr_a <= (others => '0');			             
			        t_addr_b <= (others => '0');
			             
				elsif (state_add = S_WAIT_count) then
			         if count > 1  then
			             t_addr_a <= t_addr_a +1;
			             state_add <= S_ADDING;
			             count_add <= 1;
			             mul_addr_b <= (0 => '1', others => '0');
			             mul_addr_a <= (0 => '1', others => '0');

			         end if; 			    
--===================================================			        
			    elsif (state_add =S_ADDING) then                                        
                    if (count_add mod (ramsize+1)) = ramsize-1 then
                        mul_addr_b <= mul_addr_b +1;
                        t_addr_b <= mul_addr_b;
                    else
                        t_addr_b <= t_addr_b +1;
                    end if;

                   
                    if (count_add mod (ramsize+1)) = 0  then
						mul_addr_a <= mul_addr_a +1;
                        t_addr_a <= mul_addr_a;
                    else                        
                         t_addr_a <= t_addr_a + 1;
                    end if;
										
                    if count_add = ramsize  then					
                        count_add <= count_add + 1;
					elsif count_add = (ramsize * (ramsize+1)) then 
						count_add <= 0;
						t_addr_b <= std_logic_vector(to_unsigned(ramsize-1, addr_width+1));
						t_addr_a <= (others => '0');
						mul_addr_b <= (0 => '1', others => '0');
						mul_addr_a <= (0 => '1', others => '0');
					else
						count_add <= count_add + 1;
                    end if;
                    
                    if count_add = ramsize then
                        state_add <= S_STALL_ADD_1;
                    elsif count_add = (ramsize * (ramsize+1)) then
                        state_add <= S_MUL_to_RED;
                    end if;
                                                           			    			              
			    elsif (state_add = S_STALL_ADD_1) then 
			         state_add <= S_ADDING;
			         count_add <= count_add + 1;
			         t_addr_b <= t_addr_b + 1;
			         
			         if (count_add mod (ramsize+1)) = 0  then
						mul_addr_a <= mul_addr_a +1;
                        t_addr_a <= mul_addr_a;
                    else                        
                         t_addr_a <= t_addr_a + 1;
                    end if;
--===================================================      
                
               elsif (state_add = S_MUL_to_RED) then
                    state_add <= S_RED;
                    t_addr_b <= t_addr_b+1;
--===================================================
--Reduction
--===================================================
			    elsif (state_add = S_RED) then
			       if t_addr_a = ramsize-1 then
			         state_add <=S_RED_STALL_0;
			         t_addr_a <= (others=> '0');
			         t_addr_b <= (others=> '0');
			         red_b_addr_a <= red_b_addr_a + 1;
			       else
			         t_addr_a <= t_addr_a + 1;
			         t_addr_b <= t_addr_b + 1;
			         if t_addr_a > 1 then
			             red_b_addr_a <= red_b_addr_a + 1;
			         end if;
			       end if;
                   	
                 
                 elsif (state_add = S_RED_STALL_0) then
                    if red_b_addr_a = ramsize-1 then
                        red_b_addr_a <= (others => '0');
                        red_b_addr_b <= red_b_addr_b + 1;
                        state_add <= S_RED_2;
                    else
                        red_b_addr_a <= red_b_addr_a + 1;
                        
                        if red_b_addr_a = ramsize-2 then
                            red_b_addr_b <= red_b_addr_b + 1;
                        end if;
                    end if;
                 
                 elsif (state_add = S_RED_2) then
                    if red_b_addr_b = ramsize-1 then 
                        red_b_addr_b <= (others => '0');
                        red_b_addr_a <= red_b_addr_a + 1;
                        state_add <= S_RED_STALL_1;
                    else
                        red_b_addr_b <= red_b_addr_b + 1;
                        red_b_addr_a <= red_b_addr_a + 1;
                                               
                    end if;
                    if red_add_out(15 downto 0) = x"ffff" then
                        count_all_ones <= count_all_ones + 1;
                    end if;
                    
                 elsif (state_add = S_RED_STALL_1) then
                    if red_b_addr_a = ramsize-1 then 
                        red_b_addr_a <= (others => '0');                      
                        state_add <= S_ALL_ONES;
                        if red_add_out(14 downto 0) = "11" & x"fff" then
                             count_all_ones <= count_all_ones + 1;
                        end if;
                    else
                        red_b_addr_a <= red_b_addr_a + 1;
                        if red_add_out(15 downto 0) = x"ffff" then
                            count_all_ones <= count_all_ones + 1;
                        end if; 
                    end if;
                  
                  elsif state_add = S_ALL_ONES then
                    if red_b_addr_b = ramsize-1 then 
                        red_b_addr_b <= (others => '0'); 
                        red_b_addr_a <= (others => '0');                     
                        state_add <= S_WAIT_count;
--                        state_add <= S_READ_OUT;
                        count_all_ones <= 0;
                    else
                        red_b_addr_b <= red_b_addr_b + 1; 
                    end if;
--====================================================

--        --===========================================
--        --Reading out    
--        --===========================================
--                elsif state_add = S_READ_OUT then
--                  if red_b_addr_a = ramsize-1 then
--                    red_b_addr_a <= (others=>'0');
--                    state_add <= S_WAIT_count;
--                  else
--                    red_b_addr_a <= red_b_addr_a + 1;
--                  end if;  
                
                
			end if;
		 end if;
         
          
         done_mod_mul <= done_s; 
         end if;
    end process;
    
	
    p_comb_add: process(state_add, count, t_addr_a, load_done_t, count_add,  t_addr_a, t_addr_b, red_b_addr_a, red_b_addr_b, count_all_ones)
    begin
        --! Default value
        t_wea <= '0';
        t_web <= '0';
        no_add <= '0';
        first_add <= '0';
        
        --Reduction
        red_b_wea <= '0';
        red_b_web <= '0';
        first_add_red <= '0';
        second_add_red <= '0';
        second_round <= '0';
        last_block <= '0';
        last_block_red_2 <= '0';
        
        reduction <= '0';
        
        
        valid_s <= '0';
        done_s <= '0';
        
      if (state_add = S_RESET_add) then
      
      elsif (state_add = S_WAIT_count) then
			  if count > 1  then
			     t_wea <= '1'; 
			     no_add <= '1';   
			  end if;
		   
	  elsif (state_add = S_ADDING) then
--===================================================	 
	     t_wea <= '1';
         if (count_add mod (ramsize + 1)) = 0 then
              first_add <= '1';
         end if;

         if (count_add <= ramsize) or ((count_add mod (ramsize + 1)) = (ramsize)) then
              no_add <= '1';
         end if;
--===================================================	   
	  
	  elsif (state_add = S_STALL_ADD_1) then
	      first_add <= '1';
	      
	 elsif (state_add = S_MUL_to_RED) then
	   reduction <= '1';  
	  
--===================================================	   
-- Reduction
--===================================================	   
	  
	    elsif (state_add = S_RED) then
	        if t_addr_a = 1 then
    	        first_add_red <= '1';
	        end if;
	        if t_addr_a > 1 then
    	        red_b_wea <= '1';
	        end if;	 
	        
	        reduction <= '1';  
	    
	    elsif (state_add =S_RED_STALL_0) then
	       red_b_wea <= '1';
	       if red_b_addr_a = ramsize-1 then
	           second_round <= '1';
	           second_add_red <= '1';
	       end if;
	       if red_b_addr_a = ramsize-2 then
	           last_block <= '1';     
	       end if;
	       
	       reduction <= '1'; 
	       
	    elsif (state_add = S_RED_2) then
	       red_b_wea <= '1';
	       second_round <= '1';
	       
	       reduction <= '1'; 
	    
	    elsif (state_add = S_RED_STALL_1) then
		  	red_b_wea <= '1';
	       second_round <= '1';	       
	       if red_b_addr_a = ramsize - 2 then
	           last_block_red_2 <= '1';
	       end if;
	       
	       reduction <= '1'; 
	    
	    elsif (state_add = S_ALL_ONES) then
		  	if count_all_ones = ramsize - 1 then
		  	   red_b_web <= '1';
		  	else
                red_b_web <= '0';
            end if;
            
            reduction <= '1'; 
            
            if red_b_addr_b = ramsize -1 then
                done_s <= '1';
            end if;	
        
        
        --===========================================
        --Reading out    
        --===========================================
--        elsif state_add = S_READ_OUT then
--            reduction <= '1';
--            valid_s <= '1';
--            if red_b_addr_a = ramsize -1 then
--                done_s <= '1';
--            end if;
        end if;
		end process;




--==========================Reduction================================

  t_register:process(clk)
begin
    if rising_edge(clk) then
        t_ram_out_b_reg <= t_ram_out_b(15);
    end if;
end process;



red_add_in_mux_2 <= t_ram_out_b(14 downto 0) & t_ram_out_b_reg;
red_add_in_mux_1 <= t_ram_out_a  when last_block ='0' else '0' & t_ram_out_a(14 downto 0);

red_add_in_mux_3 <= b_ram_out_b  when last_block_red_2 ='0' else '0' & b_ram_out_b(14 downto 0);


red_adder_in_1 <= zeros(47 downto 16) & red_add_in_mux_1 when second_round = '0' else (others => '0');
red_adder_in_2 <= zeros(47 downto 16) & red_add_in_mux_2 when second_round = '0' else zeros(47 downto 16) & red_add_in_mux_3;
carry_in <= '0' when first_add_red = '1' else 
            red_add_out(15) when second_add_red = '1' else 
            red_add_out(16);
--====================ADD 2=======================
trad_dsp_add2:if traditional_dsp = '1' generate 
uut_add2: add_dsp PORT MAP (
          clk => clk,
          CONCAT => red_adder_in_1,
          C => red_adder_in_2,
          CARRYIN => carry_in,
          p => red_add_out
        );          
end generate trad_dsp_add2;

infer_dsp_add2:if traditional_dsp = '0' generate 
uut_add2_2: entity work.add(behavioral) PORT MAP (
          clk => clk,
          CONCAT => red_adder_in_1,
          C => red_adder_in_2,
          CARRYIN => carry_in,
          p => red_add_out
        );          
end generate infer_dsp_add2;

 --=============================================
red_b_ram_in_a <= red_add_out(15 downto 0);



--============================== Exponentiation Logic Start ===================================
p_fsm_exp: process(clk)
    begin
        if rising_edge(clk) then
            if (rst = '1') then   
            else
				if (state_exp = S_RESET_exp) then
					state_exp           <= S_WAIT_exp;
					
				elsif (state_exp = S_WAIT_exp) then
			         if start_expo = '1' then
			             state_exp <= S_STALL_FOR_LOAD_E;
			             exp_a_addr_a <= (others => '0');
			             exp_b_addr_a <= (others => '0');
			             e_addr_a <= (others => '0');			             
			             e_addr_b <= std_logic_vector(to_unsigned(ramsize-1, e_addr_b'high + 1));
			             e_addr_count <= std_logic_vector(to_unsigned(ramsize-1, e_addr_count'high + 1));
			             count_exp <= 0;			             
			         end if;
			         
			     elsif (state_exp = S_STALL_FOR_LOAD_E) then
			         state_exp <= S_LOAD_E_OUT_exp;
			         
			     elsif (state_exp = S_LOAD_E_OUT_exp) then
			        state_exp <= S_SQUARE_exp;
			        count_exp <= 0;
			        sq_reg <= '1';
			        first_block_exp <= '1';
			        
			         
			     elsif (state_exp = S_SQUARE_exp) then
			         state_exp <= S_WAIT_SQUARE_DONE_exp;			        

			     
			     elsif (state_exp = S_WAIT_SQUARE_DONE_exp) then
			         if done_mod_mul = '1' then
                        state_exp <= S_MUL_exp;
                        sq_reg <= '0';
                      end if;
                      
                 elsif (state_exp = S_MUL_exp) then
			        state_exp <= S_WAIT_MUL_DONE_exp;
			        count_exp <= count_exp + 1;
			        if count_exp = 15 then
			             e_addr_count <= e_addr_count -1;
			             e_addr_b <=  std_logic_vector(to_unsigned(to_integer(unsigned(e_addr_b -1)) mod ramsize,  e_addr_b'high+1));
			        elsif count_exp = 14 and first_block_exp = '1' then
			             e_addr_b <=  std_logic_vector(to_unsigned(to_integer(unsigned(e_addr_b -1)) mod ramsize,  e_addr_b'high+1));
			             e_addr_count <= e_addr_count -1;
			        end if;
			     
			     elsif (state_exp = S_WAIT_MUL_DONE_exp) then
			         if done_mod_mul = '1' then
			            if e_addr_count = ones then
			               state_exp <= S_WAIT_exp;
			               e_addr_b <= std_logic_vector(to_unsigned(ramsize-1, e_addr_b'high + 1));
			               e_addr_count <= std_logic_vector(to_unsigned(ramsize-1, e_addr_count'high + 1));
			            else 
			                state_exp <= S_SQUARE_exp;
			                sq_reg    <= '1';
                            if( count_exp = 16) or (count_exp = 15 and first_block_exp = '1') then
                                count_exp <= 0;
                                first_block_exp <= '0';
                            end if;
                         end if;
                      end if;
			
			end if;
			   end if;
         end if;
    end process;
    
    
	
    p_comb_exp: process(state_exp, start_expo, e_addr_a, e_addr_b, exp_a_addr_a, exp_b_addr_a, e_reg(15), reduction, count_exp, done_mod_mul, first_block_exp)
    begin
        --! Default value
        exp_a_wea <= '0';
        exp_b_wea <= '0';
        e_wea <= '0';
        e_web <= '0';
        exp <= '0';
        exp_init <= '0';
        start_mod_mul_exp <= '0';
        e_mod_2 <= '0';
        exp_shift <= '0';
        first_exp_init <= '0';
        done_expo <= '0';
        
      if (state_exp = S_RESET_exp) then

		elsif (state_exp = S_WAIT_exp) then    
		      
		elsif (state_exp = S_LOAD_E_OUT_exp) then
		    first_exp_init <= '1';
		
		elsif (state_exp = S_SQUARE_exp) then
		     start_mod_mul_exp <= '1';
		
		elsif (state_exp = S_MUL_exp) then
		     start_mod_mul_exp <= '1';  
		
		elsif (state_exp = S_WAIT_MUL_DONE_exp) then  
		      if e_reg(15) = '0' and reduction = '1' then
		          e_mod_2 <= '1';
		      end if;
		      if done_mod_mul = '1' then
		          if count_exp =15 and first_block_exp = '1' then
		              exp_init <= '1';
		          elsif count_exp <= 15 then
    		          exp_shift <= '1';
    		      elsif count_exp = 16 then
    		          exp_init <= '1';
    		      end if;
		      end if;
		      if done_mod_mul = '1' then
			     if e_addr_count = ones then
			        done_expo <='1';
                 end if;
              end if;
		        		  	  
        end if;
		end process;



process(clk) 
begin
    if rising_edge(Clk) then
        if first_exp_init = '1' then
            e_reg <= e_ram_out_b(14 downto 0)&'0';
        elsif exp_init = '1' then
            e_reg <= e_ram_out_b; 
        elsif exp_shift = '1' then
            e_reg <= e_reg(14 downto 0)&'0';
        end if;
    end if;
end process;
--===========================Modular Exponentiation End==============================		

	
 
 
 c_ram_in_a <= c2_input;
 
 mem_c1_c2: entity work.memram(memory) 
        generic map(
                        SIZE => ramsize,  
                        ADDR_WIDTH => addr_width,
                        COL_WIDTH => 16
                    )
        port map(
                    clk => clk,
                    wea => c_wea, 
                    web => c_web,  
                    address_a => c_addr_a,
                    address_b => c_addr_b,
                    datain_a => c_ram_in_a,
                    datain_b => c_ram_in_b,
                    dataout_a => c_ram_out_a,
                    dataout_b  => c_ram_out_b         
        );

p_fsm_dec: process(clk)
    begin
        if rising_edge(clk) then
            if (rst = '1') then   
            else
				if (state_dec = S_RESET_dec) then
					state_dec           <= S_WAIT_dec;
					
				elsif (state_dec = S_WAIT_dec) then
			         if start = '1' then
			             state_dec <= S_INIT_LOAD_dec;
			             ec_addr_a <= (others => '0');
			             ec_addr_b <= (others => '0');
			             c_addr_a <= (others => '0');
			             ca_addr_b <= (others => '0');
			             ca_addr_a <= (others => '0');
			             cb_addr_b <= (others => '0');
			             cb_addr_a <= (others => '0');
			             c_addr_b <=  (others => '0');		             
			         end if;
			    
			    elsif (state_dec = S_INIT_LOAD_dec) then
			        if (c_addr_a = ramsize-1) then                           
			             state_dec <= S_C1PX;
			             c_addr_a <= (others => '0');        
			             ec_addr_b <= (others => '0'); 
			             ca_addr_b <= (others => '0');
			             cb_addr_b <= (others => '0');       
			         else
			             ec_addr_b <= ec_addr_b + 1;
			             c_addr_a <= c_addr_a + 1;
			             ca_addr_b <= ca_addr_b + 1;
			             cb_addr_b <= cb_addr_b + 1;
			         end if;
			         
			    elsif (state_dec = S_C1PX) then
			        state_dec <= S_C1PX_done;
			         
			    elsif (state_dec = S_C1PX_done) then  
			         if done_expo = '1' then
    			        state_dec <= S_MOV_B2A;
    			        cb_addr_a <= cb_addr_a + 1;
    			     end if;
    			     
    			 elsif (state_dec = S_MOV_B2A) then
    			     if cb_addr_a = ramsize -1 then 
    			         state_dec <= S_MOV_B2A_rem;
    			         cb_addr_a <= (others => '0');
    			         ca_addr_a <= ca_addr_a + 1; 
    			         ec_addr_a <= ec_addr_a + 1; 
    			         cb_addr_b <= cb_addr_b + 1;
    			     else
    			         cb_addr_a <= cb_addr_a + 1;
    			         ca_addr_a <= ca_addr_a + 1;
    			         ec_addr_a <= ec_addr_a + 1;
    			         
    			         cb_addr_b <= cb_addr_b + 1;
    			     end if;
    			 
    			 elsif (state_dec = S_MOV_B2A_rem) then
    			     if ca_addr_a = ramsize -1 then 
    			         state_dec <= S_C1PX_INV;
    			         ca_addr_a <= (others => '0');
    			         ec_addr_a <= (others => '0');
    			         cb_addr_b <= (others => '0');
    			     else
    			         ca_addr_a <= ca_addr_a + 1;
    			         ec_addr_a <= ec_addr_a + 1;
    			         cb_addr_b <= ca_addr_b + 1;
    			     end if;
    			 
    			 elsif (state_dec = S_C1PX_INV) then 
			        state_dec <= S_C1PX_INV_done;
			         
			    elsif (state_dec = S_C1PX_INV_done) then  
			         if done_expo = '1' then
    			        state_dec <= S_MOV_C2A;
    			        c_addr_b <= c_addr_b+1;
    			     end if;
    			
    			 elsif (state_dec = S_MOV_C2A) then 
    			     if c_addr_b = ramsize -1 then 
    			        state_dec <= S_MOV_C2A_rem;
    			        c_addr_b <= (others => '0');
    			        ca_addr_a <= ca_addr_a+1;
    			     else
    			        c_addr_b <= c_addr_b+1;
    			        ca_addr_a <= ca_addr_a+1;
    			     end if;
    			 
    			 elsif (state_dec = S_MOV_C2A_rem) then
    			     if ca_addr_a = ramsize -1 then 
    			         state_dec <= S_MUL_C1XC2;
    			         ca_addr_a <= (others => '0');
    			         ec_addr_a <= (others => '0');
    			         cb_addr_b <= (others => '0');
    			     else
    			         ca_addr_a <= ca_addr_a + 1;
    			     end if;
    			 
    			 elsif (state_dec = S_MUL_C1XC2) then 
			        state_dec <= S_MUL_C1XC2_done;
			         
			    elsif (state_dec =  S_MUL_C1XC2_done) then  
			         if done_mod_mul = '1' then
    			        state_dec <= S_READ_OUT_dec;
    			        cb_addr_a <= cb_addr_a + 1;
    			     end if;
    			
            --===========================================
        --Reading out    
        --===========================================
                elsif state_dec = S_READ_OUT_dec then
                  if cb_addr_a = ramsize-1 then
                    cb_addr_a <= (others=>'0');
                    state_dec <= S_WAIT_dec;
                  else
                    cb_addr_a <= cb_addr_a + 1;
                  end if;  
                       
			end if;
			   end if;
	     valid <= valid_output; 
	     done <= done_dec;
         end if;
    end process;
    
    
	
    p_comb_dec: process(state_dec, start, c_addr_a, c_addr_b, cb_addr_a, done_expo, done_mod_mul)
    begin
        --! Default value
        dec <= '0';
        c_wea <= '0';
        c_web <= '0';
        ec_wea <= '0';
        ec_web <= '0';
        ca_web <= '0';
        ca_wea <= '0';
        cb_web <= '0';
        cb_wea <= '0';
        move_b2a <= '0';
        sel_c2a <= '0';
        start_expo <= '0';
        start_mul_dec <= '0';
        
        valid_output <= '0';
        done_dec <= '0';
        
      if (state_dec = S_RESET_dec) then

		elsif (state_dec = S_WAIT_dec) then
		    if start = '1' then
		      dec <= '1';
		    end if;
		
		elsif (state_dec = S_INIT_LOAD_dec) then
		    dec <= '1';
		    c_wea <= '1';
		    ec_web <= '1';
		    ca_web <= '1';
            cb_web <= '1';
        
        elsif (state_dec = S_C1PX) then
            start_expo <= '1';
		
		elsif (state_dec = S_C1PX_done) then  
		  if done_expo = '1' then
		      move_b2a <= '1';
		  end if;
		
		elsif (state_dec = S_MOV_B2A) then
		   move_b2a <= '1';
		   ca_wea <= '1';
		   ec_wea <= '1';
		   cb_web <= '1';
		
		elsif (state_dec = S_MOV_B2A_rem) then
		   move_b2a <= '1';
		   ca_wea <= '1';
		   ec_wea <= '1';
		   cb_web <= '1';
		
		elsif (state_dec = S_C1PX_INV) then
            start_expo <= '1';
		
		elsif (state_dec = S_C1PX_INV_done) then  
		  if done_expo = '1' then
            move_b2a <= '1';
            sel_c2a <= '1';
		  end if;
		
		elsif (state_dec = S_MOV_C2A) then
		   move_b2a <= '1';
		   ca_wea <= '1';
		   sel_c2a <= '1';
		
		elsif (state_dec = S_MOV_C2A_rem) then
		   move_b2a <= '1';
		   ca_wea <= '1';
		   sel_c2a <= '1';
		   
		elsif (state_dec = S_MUL_C1XC2) then 
		  start_mul_dec <= '1';
		
		elsif (state_dec = S_MUL_C1XC2_done) then 
		      if done_mod_mul = '1' then
		          move_b2a <= '1';
		          valid_output <= '1'; 
		      end if;
		
		elsif state_dec = S_READ_OUT_dec then
		      move_b2a <= '1';
		      valid_output <= '1';   
		      if cb_addr_a = ramsize-1 then
		          done_dec <= '1';
		      end if;   
		        		  	  
        end if;
		end process;
--===========================end==============================
end Behavioral;
