----------------------------------------------------------------------------------
-- This file is dual port block RAM 

-- Copyright (C) 2021
-- Authors: Sanjay Deshpande <sanjay.deshpande@yale.edu>
          
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 3 of the License, or
-- (at your option) any later version.

-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.

-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software Foundation,
-- Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
----------------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;
use std.textio.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_textio.all;


entity memram is
  generic (
    SIZE       : integer := 4;
    ADDR_WIDTH : integer := 2;
    COL_WIDTH  : integer := 256;
    InitFile   : string  := ""
    );
  port(
    clk  : in  std_logic;
    wea   : in  std_logic;
    web   : in  std_logic;
    address_a : in  std_logic_vector(ADDR_WIDTH-1 downto 0);
    address_b : in  std_logic_vector(ADDR_WIDTH-1 downto 0);
    datain_a   : in  std_logic_vector(COL_WIDTH-1 downto 0);
    datain_b   : in  std_logic_vector(COL_WIDTH-1 downto 0);
    dataout_a   : out std_logic_vector(COL_WIDTH-1 downto 0) := (others => '0');
    dataout_b   : out std_logic_vector(COL_WIDTH-1 downto 0) := (others => '0')
    );
end memram ;

architecture memory of memram  is
--  type ram_type is array (SIZE-1 downto 0) of std_logic_vector(COL_WIDTH-1 downto 0);
  type ram_type is array (0 to SIZE - 1) of std_logic_vector(COL_WIDTH-1 downto 0);


  -- Initializes the RAM from a file
  impure function InitRamFromFile (RamFileName : in string) return Ram_Type is
    file RamFile         : text is in RamFileName;
    variable RamFileLine : line;
    variable RAM         : Ram_Type;
  begin
    --Default value 
    for I in Ram_Type'range loop
      readline (RamFile, RamFileLine);
      read (RamFileLine, RAM(I));
    end loop;

    return RAM;
  end function;


  --Wrapper for the RAM initialization. In case of empty string the RAM is
  --initialized with zeros. Necesarry because declaration in function above
  --wants to open a ffile directy (mismatch in behavior modelsim vs xst)
  impure function default_init (RamFileName : in string) return Ram_Type is
    variable RAM      : Ram_Type;
    variable comp_val : string(1 to RamFileName'length);
  begin
    --Default value
    
    if RamFileName = "" then
      --report "Initializing" severity failure;
      
      for I in Ram_Type'range loop
        RAM(I) := (others => '0');
      end loop;
    else
      --Initialize
      RAM := InitRamFromFile(InitFile);
    end if;

    return RAM;
  end function;

  shared variable RAM : ram_type := default_init(InitFile);

  -- output registers
  signal address_a_reg : std_logic_vector(ADDR_WIDTH-1 downto 0);
  signal address_b_reg : std_logic_vector(ADDR_WIDTH-1 downto 0);
  
begin

  process (clk)
  begin
    if rising_edge(clk) then
      if wea = '1' then
        ram(conv_integer(address_a)) := datain_a;
      end if;
        dataout_a <= ram(conv_integer(address_a));
    end if;
  end process;

  process (clk)
  begin
    if rising_edge(clk) then
      if web = '1' then
        ram(conv_integer(address_b)) := datain_b;      
      end if;
        dataout_b <= ram(conv_integer(address_b));  
    end if;
  end process;




end memory;


