-------------------------------------------------------------------------------------
-- This file is modular reduction unit targeted for prime 2^1279 - 1 

-- Copyright (C) 2021
-- Authors: Sanjay Deshpande <sanjay.deshpande@yale.edu>
          
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 3 of the License, or
-- (at your option) any later version.

-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.

-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software Foundation,
-- Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA


--This module takes 2*1279 bits as input and reduces them to 1279 bits
-------------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx leaf cells in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity reduction_1279_3 is
generic (

            traditional_dsp : std_logic := '1'; 
            ramsize : integer := 80;
            addr_width : integer := 7;
            MSB : integer := 23

        );
Port ( 
           clk   : in STD_LOGIC;
           start : in STD_LOGIC;
           rst   : in STD_LOGIC;
           done  : out STD_LOGIC;
           valid  : out STD_LOGIC;
           a_input : in STD_LOGIC_VECTOR (15 downto 0);
           output: out STD_LOGIC_VECTOR (15 downto 0)

           );
end reduction_1279_3;

architecture Behavioral of reduction_1279_3 is
    COMPONENT add_dsp
    PORT(
         clk : in STD_LOGIC;
           CONCAT : in STD_LOGIC_VECTOR (47 downto 0);
           C : in STD_LOGIC_VECTOR (47 downto 0);
           p : out STD_LOGIC_VECTOR (47 downto 0);
           CARRYIN : in STD_LOGIC  
        );
    END COMPONENT;
signal a_addr_a, a_addr_b : std_logic_vector(addr_width+1 downto 0) := (others => '0');
signal a_wea, a_web: std_logic := '0';
signal a_ram_in_a, a_ram_in_b : std_logic_vector(15 downto 0);    
signal a_ram_out_a, a_ram_out_b : std_logic_vector(15 downto 0); 
signal a_ram_out_b_reg : std_logic_vector(15 downto 0); 
--signal a_ram_out_b_reg : std_logic; 

signal b_addr_a, b_addr_b : std_logic_vector(addr_width-1 downto 0) := (others => '0');
signal b_wea, b_web: std_logic := '0';
signal b_ram_in_a, b_ram_in_b : std_logic_vector(15 downto 0);    
signal b_ram_out_a, b_ram_out_b : std_logic_vector(15 downto 0); 

signal add_in_mux_1 : std_logic_vector(15 downto 0);
signal add_in_mux_2 : std_logic_vector(15 downto 0);
signal add_in_mux_3 : std_logic_vector(15 downto 0);

signal adder_in_1 : std_logic_vector(47 downto 0) := (others => '0'); 
signal adder_in_2 : std_logic_vector(47 downto 0) := (others => '0'); 
signal add_out : std_logic_vector(47 downto 0) := (others => '0'); 

signal second_round : std_logic := '0'; 
signal first_add_red : std_logic := '0'; 
signal second_add_red : std_logic := '0'; 

signal carry_in : std_logic := '0'; 
signal last_block, last_block_red_2 : std_logic := '0';
signal third_block, third_last_block : std_logic := '0';

signal negative: std_logic := '0';

constant zeros : std_logic_vector(47 downto 0) := (others => '0'); 

type state_s is (S_RESET, S_WAIT_START, S_STALL_0, S_INIT_LOAD, S_LOAD, S_STALL_1, S_RED, S_RED_2, S_STALL_2, S_ALL_ONES, S_RED_3, S_STALL_3, S_RED_4, S_STALL_4,   S_DONE, S_MINUS_SCANNING, S_MINUS_STALL, S_READ_OUT);
signal state        			: state_s;  

signal count_all_ones : integer range 0 to ramsize;


signal valid_S, done_s: std_logic := '0';

 

signal borrow: std_logic := '0';
signal minus: std_logic := '0';
signal borrow_in, borrow_out : std_logic := '0';
signal sub_in1, sub_in2 : std_logic_vector(15 downto 0) := (others=> '0');
signal sub_out : std_logic_vector(15 downto 0) := (others=> '0');

signal b_minus_neg, b_minus_borrow  : std_logic_vector(16 downto 0) := (others=> '0');

begin

a_ram_in_a <= a_input and (x"1fff") when a_addr_a = 239 else a_input;

b_minus_neg <= '0'&b_ram_out_a - negative;

b_minus_borrow <= '0'&b_ram_out_a - borrow;

output <=  b_minus_neg(15 downto 0) when b_addr_a = 1 else b_minus_borrow(15 downto 0);
--output <=  b_ram_out_a - negative when b_addr_a = 1 else b_ram_out_a ;

process(Clk)
begin
    if rising_edge(Clk) then
        if b_addr_a = 0 then 
            borrow <= '0';
        elsif b_addr_a = 1 then 
            borrow <= b_minus_neg(16);
        else
            borrow <= b_minus_borrow(16);
        end if;
    end if;
end process;

            

	mem_a: entity work.memram(memory) 
            generic map(
                            SIZE => 3*ramsize,
                            ADDR_WIDTH => addr_width+1+1,
                            COL_WIDTH => 16
                        )
            port map(
                        clk => clk,
                        wea => a_wea, 
                        web => a_web,  
                        address_a => a_addr_a,
                        address_b => a_addr_b,
                        datain_a => a_ram_in_a,
                        datain_b => a_ram_in_b,
                        dataout_a => a_ram_out_a,
                        dataout_b  => a_ram_out_b         
            );
  
  
  p_register:process(clk)
begin
    if rising_edge(clk) then
        a_ram_out_b_reg <= a_ram_out_b;
    end if;
end process;


add_in_mux_2 <= '0'&a_ram_out_b(13 downto 0) & a_ram_out_b_reg(15) when last_block ='1' else  
                '0' & (a_ram_out_b(12 downto 0)) & (a_ram_out_b_reg(15 downto 14)) when third_last_block ='1' else  
                 a_ram_out_b(13 downto 0) & a_ram_out_b_reg(15 downto 14) when third_block ='1' else 
                 a_ram_out_b(14 downto 0) & a_ram_out_b_reg(15);


add_in_mux_1 <= '0' & a_ram_out_a(14 downto 0)  when last_block ='1' else 
                '0' & b_ram_out_b(14 downto 0)  when third_last_block ='1' else 
                b_ram_out_b when third_block = '1' else
                a_ram_out_a;
                
add_in_mux_3 <= b_ram_out_b  when last_block_red_2 ='0' else '0' & b_ram_out_b(14 downto 0);

--add_in_mux_3 <= '0' & b_ram_out_b(14 downto 0) when last_block_red_2 ='1' else 
--                b_ram_out_b;


adder_in_1 <= zeros(47 downto 16) & add_in_mux_1 when second_round = '0' else (others => '0');
adder_in_2 <= zeros(47 downto 16) & add_in_mux_2 when second_round = '0' else zeros(47 downto 16) & add_in_mux_3;
carry_in <= '0' when first_add_red = '1' else 
            add_out(15) when second_add_red = '1' else 
            add_out(16);

       
 
 
 --=========================ADD==================================
trad_dsp_add:if traditional_dsp = '1' generate 
uut_add: add_dsp PORT MAP (
          clk => clk,
          CONCAT => adder_in_1,
          C => adder_in_2,
          CARRYIN => carry_in,
          p => add_out
        );   
end generate trad_dsp_add;

infer_dsp_add:if traditional_dsp = '0' generate 
uut_add: entity work.add(behavioral) PORT MAP (
          clk => clk,
          CONCAT => adder_in_1,
          C => adder_in_2,
          CARRYIN => carry_in,
          p => add_out
        );
end generate infer_dsp_add;
--==============================================================

b_ram_in_a <= add_out(15 downto 0);

b_ram_in_b <= (others => '0');
	mem_b: entity work.memram(memory) 
            generic map(
                            SIZE => ramsize,
                            ADDR_WIDTH => addr_width,
                            COL_WIDTH => 16
                        )
            port map(
                        clk => clk,
                        wea => b_wea, 
                        web => b_web,  
                        address_a => b_addr_a,
                        address_b => b_addr_b,
                        datain_a => b_ram_in_a,
                        datain_b => b_ram_in_b,
                        dataout_a => b_ram_out_a,
                        dataout_b  => b_ram_out_b         
            );
    
  p_fsm: process(clk)
    begin
        if rising_edge(clk) then
            if (rst = '1') then  
			
            else
				if (state = S_RESET) then
					state           <= S_WAIT_START;
					
				elsif (state = S_WAIT_START) then
			         if start = '1' then
			             state <= S_INIT_LOAD;
			             a_addr_a <= (others => '0');
			             a_addr_b <= (others => '0');			             
			             b_addr_a <= (others => '0');
			             b_addr_b <= (others => '0');
			             count_all_ones <= 0;
			         end if;
			    
			            
			    elsif (state = S_INIT_LOAD) then
			         if (a_addr_a = 3*ramsize-1) then                           
			             state <= S_RED;
			             a_addr_a <= (others => '0');
			             a_addr_b <= a_addr_b +1; 
			             
			             if a_addr_a = MSB then
			                 negative <= a_input(a_input'high);
			             end if; 
			            
			         else
			             a_addr_a <= a_addr_a + 1;	
			             a_addr_b <= std_logic_vector(to_unsigned(ramsize-1, addr_width+1+1)); -- x"0";
			             
			             if a_addr_a = MSB then
			                 negative <= a_input(a_input'high);
			             end if; 	
			         end if;
			   
			   

			    elsif (state = S_RED) then
			       if a_addr_a = ramsize-1 then
			         state <=S_STALL_0;
			         a_addr_a <= (others=> '0');
			         a_addr_b <= (others=> '0');
			         b_addr_a <= b_addr_a + 1;
			       else
			         a_addr_a <= a_addr_a + 1;
			         a_addr_b <= a_addr_b + 1;
			         if a_addr_a > 1 then
			             b_addr_a <= b_addr_a + 1;
			         end if;
			       end if;
                   	
                 
                 elsif (state = S_STALL_0) then
                    if b_addr_a = ramsize-1 then
                        b_addr_a <= (others => '0');
                        b_addr_b <= b_addr_b + 1;
                        state <= S_RED_2;
                    else
                        b_addr_a <= b_addr_a + 1;
                        
                        if b_addr_a = ramsize-2 then
                            b_addr_b <= b_addr_b + 1;
                        end if;
                    end if;
                 
                 elsif (state = S_RED_2) then
                    if b_addr_b = ramsize-1 then 
                        a_addr_b <= std_logic_vector(to_unsigned(2*ramsize-1, addr_width+1+1)); --'0'&x"f";
                        b_addr_b <= (others => '0');
                        b_addr_a <= b_addr_a + 1;
                        state <= S_STALL_1;
                    else
                        b_addr_b <= b_addr_b + 1;
                        b_addr_a <= b_addr_a + 1;
                                               
                    end if;

                    
                 elsif (state = S_STALL_1) then
                    if b_addr_a = ramsize-1 then 
                        a_addr_b <= a_addr_b+1;
                        b_addr_a <= (others => '0');
                        state <= S_RED_3;
                    else
                        b_addr_a <= b_addr_a + 1;
                    end if;
                  
                  
                  elsif (state = S_RED_3) then
                    if b_addr_b = ramsize - 1 then
                       b_addr_b <= (others => '0');
                       a_addr_b <= (others => '0');
                       b_addr_a <= b_addr_a + 1;
                       state <= S_STALL_3;
                    else
                       b_addr_b <= b_addr_b +1;
                       a_addr_b <= a_addr_b + 1; 
                       if b_addr_b > 1 then
			             b_addr_a <= b_addr_a + 1;
			         end if;
                    end if;
                  
                  elsif (state = S_STALL_3) then
                    if b_addr_a = ramsize-1 then
                        b_addr_a <= (others => '0');
                        b_addr_b <= b_addr_b + 1;
                        state <= S_RED_4;
                    else
                        b_addr_a <= b_addr_a + 1;
                        
                        if b_addr_a = ramsize-2 then
                            b_addr_b <= b_addr_b + 1;
                        end if;
                    end if;
 
                  elsif (state = S_RED_4) then
                    if b_addr_b = ramsize-1 then 
                        b_addr_b <= (others => '0');
                        b_addr_a <= b_addr_a + 1;
                        state <= S_STALL_4;
                    else
                        b_addr_b <= b_addr_b + 1;
                        b_addr_a <= b_addr_a + 1;
                                               
                    end if;
                    if add_out(15 downto 0) = x"ffff" then
                        count_all_ones <= count_all_ones + 1;
                    end if;

                 elsif (state = S_STALL_4) then
                    if b_addr_a = ramsize-1 then 
                        a_addr_b <= a_addr_b+1;
                        b_addr_a <= (others => '0');                      
                        state <= S_ALL_ONES;
                        if add_out(14 downto 0) = "11" & x"fff" then
                             count_all_ones <= count_all_ones + 1;
                        end if;
                    else
                        b_addr_a <= b_addr_a + 1;
                        if add_out(15 downto 0) = x"ffff" then
                            count_all_ones <= count_all_ones + 1;
                        end if; 
                    end if;
                    
                  elsif state = S_ALL_ONES then
                    if b_addr_b = ramsize-1 then 
                        b_addr_b <= (others => '0');                      
                        state <= S_READ_OUT;
                        count_all_ones <= 0;
                        valid_s <= '1';
                    else
                        b_addr_b <= b_addr_b + 1; 
                    end if;
                  
        --===========================================
        --Reading out    
        --===========================================
                elsif state = S_READ_OUT then
                  if b_addr_a = ramsize-1 then
                     b_addr_a <= (others=>'0');
                    state <= S_WAIT_START;
                    valid_S <= '0';
                  else
                     b_addr_a <= b_addr_a + 1;
                     valid_s <= '1';
                  end if; 
                  
                  
                 end if;
                  			
			end if; 
			valid <= valid_s; 
			done <= done_s;             
         end if;
    end process;
    
	
    p_comb: process(state, start, a_addr_a, a_addr_b, b_addr_a, b_addr_b, count_all_ones)
    begin
        --! Default value
        a_wea <= '0';
        a_web <= '0';
        b_wea <= '0';
        b_web <= '0';
        first_add_red <= '0';
        second_add_red <= '0';
        second_round <= '0';
        last_block <= '0';
        last_block_red_2 <= '0';
        third_block <= '0';
        third_last_block <= '0';
        minus <= '0';
        done_s <= '0';
        
      if (state = S_RESET) then

		elsif (state = S_WAIT_START) then

		    		
		elsif (state = S_INIT_LOAD) then
	        a_wea <= '1';
	    
	    elsif (state = S_MINUS_SCANNING) then
	        a_wea <= '1';
	        minus <= '1';
	    
	    elsif (state = S_MINUS_STALL) then
	        a_wea <= '1';
	        minus <= '1';
	    
	    elsif (state = S_RED) then
	        if a_addr_a = 1 then
    	        first_add_red <= '1';
	        end if;
	        if a_addr_a > 1 then
    	        b_wea <= '1';
	        end if;	   
	    
	    elsif (state =S_STALL_0) then
	       b_wea <= '1';
	       if b_addr_a = ramsize-1 then
	           second_round <= '1';
	           second_add_red <= '1';
	       end if;
	       if b_addr_a = ramsize-2 then
	           last_block <= '1';
	           
	       end if;
	       
	    elsif (state = S_RED_2) then
	       b_wea <= '1';
	       second_round <= '1';
	    
	    elsif (state = S_STALL_1) then
		  	b_wea <= '1';
	       second_round <= '1';	       
	       if b_addr_a = ramsize - 2 then
	           last_block_red_2 <= '1';
	       end if;
	    
	    elsif (state = S_RED_3) then
	       third_block <= '1';
	       if b_addr_b = 1 then
    	        first_add_red <= '1';
	        end if;
	        if b_addr_b > 1 then
	           b_wea <= '1';
	        end if;
	    
	    elsif (state =S_STALL_3) then
	       b_wea <= '1';
	       third_block <= '1';
	       if b_addr_a = ramsize-2 then
	           third_last_block <= '1'; 
	       end if;
	       
	       if b_addr_a = ramsize-1 then
	           second_round <= '1';
	           second_add_red <= '1';
	       end if;
	    
	    elsif (state = S_RED_4) then
	       b_wea <= '1';
	       second_round <= '1';
	    
	    elsif (state = S_STALL_4) then
		  	b_wea <= '1';
	       second_round <= '1';	       
	       if b_addr_a = ramsize - 2 then
	           last_block_red_2 <= '1';
	       end if;
	       
	    elsif (state = S_ALL_ONES) then
		  	if count_all_ones = ramsize - 1 then
		  	   b_web <= '1';
		  	   done_s <= '1';
		  	else
                b_web <= '0';
            end if;	
                     
        end if;
		end process;
		
		

end Behavioral;
